"""
    pyneovi is a Python wrapper around the API provided by Intrepid Control
    Systems for communicating with their NeoVI range of devices.

    Copyright © 2015-2018 John Kemp

    Web:   http://kempj.co.uk
    
    Email: kemp@kempj.co.uk

    pyneovi is provided under the MIT License terms. For license information see LICENSE.txt.
"""
